import time
import uuid

from flask import g, request
from typing import Any

from nsj_audit_lib.dto.dto_auditavel import DTOAuditavel
from nsj_audit_lib.settings import DATABASE_NAME
from nsj_audit_lib.util.audit_config import AuditConfig
from nsj_audit_lib.util.audit_ids_util import (
    retrieve_area_atendimento_id,
    retrieve_grupo_empresarial_id,
    retrieve_tenant_id,
)
from nsj_audit_lib.util.audit_response_util import (
    extract_error_info,
    normalize_response_body,
    sanitize_payload,
    unpack_response,
)
from nsj_audit_lib.util.audit_util import AuditUtil, DBTypes, HTTPMethods
from nsj_audit_lib.util.user_audit_util import get_actor_user_id, get_db_user
from nsj_audit_lib.util.util_normaliza_parametros import get_params_normalizados


class AuditRequestUtil:
    def __init__(
        self, audit_config: AuditConfig, dto_class: type[DTOAuditavel]
    ) -> None:
        self.audit_config = audit_config
        self.dto_class = dto_class

    def record_audit_request(self, **kwargs):
        # Gerando ID da requisição
        request_id = uuid.uuid4()
        g.request_id = request_id

        # Recuperando IDs de tenant, grupo_empresarial e area_atendimento
        query_args = request.args
        body = request.get_json(force=True, silent=True)
        tenant_id = retrieve_tenant_id(
            self.audit_config,
            query_args=query_args,
            body=body,
        )
        grupo_empresarial_id = retrieve_grupo_empresarial_id(
            self.audit_config,
            query_args=query_args,
            body=body,
        )
        area_atendimento_id = retrieve_area_atendimento_id(
            self.audit_config,
            query_args=query_args,
            body=body,
        )
        g.audit_request_start_ts = time.perf_counter()
        g.audit_tenant_id = tenant_id
        g.audit_grupo_empresarial_id = grupo_empresarial_id
        g.audit_area_atendimento_id = area_atendimento_id

        # Montando o conteúdo truncado do body da requisição
        raw_body = request.get_data(cache=True, as_text=True) or None
        if raw_body:
            sanitized_body = sanitize_payload(raw_body)
            request_raw_truncated = sanitized_body[:255]
        else:
            request_raw_truncated = None

        # Montando os parâmetros normalizados
        params_normalizados = get_params_normalizados(
            query_args=query_args,
            body=body,
            dto_class=self.dto_class,
            path_args=kwargs,
        )
        g.audit_params_normalizados = params_normalizados

        # Chamando auditoria de eventos (AuditUtil.emit_request_started)
        audit_util = AuditUtil()
        audit_util.emit_request_started(
            request_id=request_id,
            tenant_id=tenant_id,
            grupo_empresarial_id=grupo_empresarial_id,
            area_atendimento_id=area_atendimento_id,
            db_type=(
                self.audit_config.db_type
                if self.audit_config and self.audit_config.db_type is not None
                else (
                    DBTypes.MULTIBANCO
                    if getattr(g, "external_database", None) is not None
                    else DBTypes.OTHER
                )
            ),
            db_key=(
                getattr(g, "external_database", {}).get("name")
                if getattr(g, "external_database", None)
                else DATABASE_NAME
            ),
            db_user=get_db_user(),
            actor_user_id=get_actor_user_id(),
            subject_user_id=None,
            session_id=None,
            http_method=HTTPMethods.from_value(request.method),
            http_route=request.path,
            action="route_called",
            params_normalizados=params_normalizados,
            request_raw_truncated=request_raw_truncated,
        )

    def record_audit_response(self, response: Any):
        request_id = getattr(g, "request_id", None) or getattr(self, "request_id", None)
        if request_id is None:
            return

        start_ts = getattr(g, "audit_request_start_ts", None)
        duration_ms = int((time.perf_counter() - start_ts) * 1000) if start_ts else 0

        body, http_status = unpack_response(response)
        body_text, body_json = normalize_response_body(body)

        is_transaction_intent = request.method in {"POST", "PUT", "PATCH", "DELETE"}

        error_code, error_message_short, error_fingerprint = extract_error_info(
            http_status,
            body_json,
            body_text,
        )
        error_normalized = None
        if error_code or error_message_short or error_fingerprint:
            error_normalized = {
                "error_code": error_code,
                "error_message_short": error_message_short,
                "error_fingerprint": error_fingerprint,
            }

        request_json = None
        if http_status >= 400:
            request_json = request.get_data(cache=True, as_text=True) or None
            if request_json:
                request_json = sanitize_payload(request_json)
                if len(request_json) > 4096:
                    request_json = request_json[:4096]

        audit_util = AuditUtil()
        audit_util.emit_request_finished(
            request_id=request_id,
            tenant_id=getattr(g, "audit_tenant_id", None),
            grupo_empresarial_id=getattr(g, "audit_grupo_empresarial_id", None),
            area_atendimento_id=getattr(g, "audit_area_atendimento_id", None),
            db_user=get_db_user(),
            http_status=http_status,
            duration_ms=duration_ms,
            tx_attempted=bool(is_transaction_intent),
            error_normalized=error_normalized,
            error_code=error_code,
            error_message_short=error_message_short,
            error_fingerprint=error_fingerprint,
            request_json=request_json,
            is_transaction_intent=is_transaction_intent,
        )



from nsj_audit_lib.dto.dto_auditavel import DTOAuditavel


def _normalize_query_args(query_args):
    if query_args is None:
        return {}
    if hasattr(query_args, "to_dict"):
        return query_args.to_dict(flat=True)
    return dict(query_args)


def get_params_normalizados(
    query_args, body, dto_class: type[DTOAuditavel] | None, path_args=None
):
    """
    Normaliza os parâmetros da requisição para auditoria.

    - query_args: parâmetros da query string.
    - body: corpo da requisição (dict ou lista de dicts).
    - dto_class: DTO associado à rota, usado para selecionar campos resumo.
    - path_args: argumentos da rota (kwargs).
    """
    normalized_query_args = _normalize_query_args(query_args)

    if dto_class is None or body is None or body == {}:
        normalized_body = {}
    else:
        normalized_body = dto_class.get_params_normalizados(body, query_args, path_args)

    return {
        "query_args": normalized_query_args,
        "body": normalized_body,
        "path_args": path_args or {},
    }



from dataclasses import dataclass

from nsj_audit_lib.util.audit_util import DBTypes


@dataclass(frozen=True, slots=True)
class AuditConfig:
    tenant_field: str | None = None
    grupo_empresarial_field: str | None = None
    area_atendimento_field: str | None = None
    db_type: DBTypes | None = None



from redis import Redis

from nsj_audit_lib.settings import AUDIT_REDIS_URL, get_logger

redis_client = None
if AUDIT_REDIS_URL:
    redis_client = Redis.from_url(
        AUDIT_REDIS_URL,
        socket_connect_timeout=0.05,
        socket_timeout=0.05,
    )
else:
    get_logger().warning(
        "Atenção! Redis não configurado (faltando variável AUDIT_REDIS_URL)."
    )



import logging
import os

APP_NAME = os.getenv("APP_NAME", "nsj_audit_lib")


def get_logger():
    return logging.getLogger(APP_NAME)


DATABASE_NAME = os.getenv("DATABASE_NAME", "")
DATABASE_USER = os.getenv("DATABASE_USER", "")
DATABASE_DRIVER = os.getenv("DATABASE_DRIVER", "POSTGRES")

AUDIT_REDIS_URL = os.getenv("AUDIT_REDIS_URL")
AUDIT_STREAM_KEY = os.getenv("AUDIT_STREAM_KEY", "audit:requests")
AUDIT_OUTBOX_TRANSACTION = (
    os.getenv("AUDIT_OUTBOX_TRANSACTION", "true").lower() == "true"
)



import uuid

from typing import Any


def is_id_segment(segment: str) -> bool:
    if segment.isdigit():
        return True
    try:
        uuid.UUID(segment)
        return True
    except (TypeError, ValueError):
        return False


def coerce_int(value: Any) -> int | None:
    if value is None:
        return None
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def coerce_uuid(value: Any, allow_none: bool = False) -> str | None:
    if value is None:
        return None if allow_none else str(uuid.UUID(int=0))
    try:
        return str(uuid.UUID(str(value)))
    except (TypeError, ValueError):
        return None if allow_none else str(uuid.UUID(int=0))



from flask import request

from nsj_audit_lib.util.audit_config import AuditConfig


def _retrieve_id_from_request(
    keys: list[str],
    custom_field_name: str | None,
    query_args=None,
    body=None,
):
    search_keys = keys
    if custom_field_name and custom_field_name not in search_keys:
        search_keys = [custom_field_name, *search_keys]

    if query_args is None:
        query_args = request.args

    for key in search_keys:
        value = query_args.get(key)
        if value is not None:
            return value

    if body is None:
        body = request.get_json(force=True, silent=True)

    if isinstance(body, dict):
        for key in search_keys:
            value = body.get(key)
            if value is not None:
                return value
    elif isinstance(body, list):
        for item in body:
            if not isinstance(item, dict):
                continue
            for key in search_keys:
                value = item.get(key)
                if value is not None:
                    return value

    return None


def _get_audit_config_field(audit_config: AuditConfig | None, field_name: str):
    if audit_config is None:
        return None
    return getattr(audit_config, field_name, None)


def retrieve_tenant_id(audit_config: AuditConfig | None, query_args=None, body=None):
    """
    Recupera o tenant_id dos parâmetros da requisição, ou do corpo da requisição e o retorna.
    """

    tenant_keys = ["tenant", "tenant_id", "id_tenant", "tenantid", "idtenant"]
    return _retrieve_id_from_request(
        tenant_keys,
        _get_audit_config_field(audit_config, "tenant_field"),
        query_args=query_args,
        body=body,
    )


def retrieve_grupo_empresarial_id(
    audit_config: AuditConfig | None,
    query_args=None,
    body=None,
):
    """
    Recupera o grupo_empresarial_id dos parâmetros da requisição, ou do corpo da requisição e o retorna.
    """

    grupo_empresarial_keys = [
        "grupo_empresarial",
        "grupo_empresarial_id",
        "id_grupo_empresarial",
        "grupoempresarial",
        "grupoempresarial_id",
        "idgrupoempresarial",
    ]
    return _retrieve_id_from_request(
        grupo_empresarial_keys,
        _get_audit_config_field(audit_config, "grupo_empresarial_field"),
        query_args=query_args,
        body=body,
    )


def retrieve_area_atendimento_id(
    audit_config: AuditConfig | None,
    query_args=None,
    body=None,
):
    """
    Recupera o area_atendimento_id dos parâmetros da requisição, ou do corpo da requisição e o retorna.
    """

    area_atendimento_keys = [
        "area_atendimento",
        "area_atendimento_id",
        "id_area_atendimento",
        "areaatendimento",
        "areaatendimento_id",
        "idareaatendimento",
    ]
    return _retrieve_id_from_request(
        area_atendimento_keys,
        _get_audit_config_field(audit_config, "area_atendimento_field"),
        query_args=query_args,
        body=body,
    )



import enum
import json
import redis
import uuid

from datetime import datetime, timezone

from nsj_audit_lib.redis_config import redis_client
from nsj_audit_lib.settings import AUDIT_STREAM_KEY, get_logger


class DBTypes(enum.Enum):
    MULTIBANCO = "mb"
    WEB_MULTITENANT = "dbprod"
    DIARIO_UNICO = "dbdu"
    OTHER = "other"
    UNDEFINED = "undefined"


class HTTPMethods(enum.Enum):
    GET = "GET"
    LIST = "LIST"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    OPTIONS = "OPTIONS"

    @staticmethod
    def from_value(value: str) -> "HTTPMethods":
        for method in HTTPMethods:
            if method.value == value:
                return method
        raise ValueError(f"Invalid HTTP method value: {value}")


class AuditUtil:

    def __init__(
        self,
        redis_client: redis.Redis = redis_client,
        audit_stream_key: str = AUDIT_STREAM_KEY,
    ) -> None:
        self.redis_client = redis_client
        self.stream_key = audit_stream_key

    def emit_request_started(
        self,
        request_id: uuid.UUID,
        tenant_id: int,
        grupo_empresarial_id: uuid.UUID | None,
        area_atendimento_id: uuid.UUID | None,
        db_type: DBTypes,
        db_key: str,
        db_user: str,
        actor_user_id: str,
        subject_user_id: int | None,
        session_id: str | None,
        http_method: HTTPMethods,
        http_route: str,
        action: str,
        params_normalizados: dict,
        meta_human: str | None = None,
        request_raw_truncated: str | None = None,
        payload_ref: str | None = None,
    ) -> str:
        """
        Publica um evento 'request_started' no Redis Stream.

        Parâmetros:
        - request_id: ID da requisição
        - tenant_id: ID do tenant
        - grupo_empresarial_id: ID do grupo empresarial
        - area_atendimento_id: ID da area de atendimento
        - db_type: Tipo do banco de dados
        - db_key: Chave de identificação do banco de dados (pode ser o nome do banco)
        - db_user: Nome do usuário do banco
        - actor_user_id: ID do usuário que fez a requisição
        - subject_user_id: ID do usuário personificado nessa requisição (se houver personificação)
        - session_id: ID da sessão (se houver, e para as APIs stateless normalmente não há)
        - http_method: Método HTTP da requisição
        - http_route: Rota HTTP da requisição (endpoint)
        - action: Ação da requisição
        - params_normalizados: Parâmetros da requisição normalizados (não é a requisição completa)
        - meta_human: Metadadtos, humanizados, da requisição
        - request_raw_truncated: Conteúdo da requisição truncado
        - payload_ref: URL da requisição gravada num bucket (se houver)
        """

        try:
            now = datetime.now(timezone.utc)

            event = {
                "event_id": str(uuid.uuid4()),
                "event_ts_utc": now.isoformat(timespec="milliseconds"),
                "event_type": "request_started",
                "request_id": str(request_id),
                "tenant_id": str(tenant_id),
                "grupo_empresarial_id": (
                    "" if grupo_empresarial_id is None else str(grupo_empresarial_id)
                ),
                "area_atendimento_id": (
                    "" if area_atendimento_id is None else str(area_atendimento_id)
                ),
                "db_key": f"{db_type.value}_{db_key}",
                "db_user": db_user,
                "actor_user_id": str(actor_user_id),
                "subject_user_id": (
                    "" if subject_user_id is None else str(subject_user_id)
                ),
                "session_id": "" if session_id is None else session_id,
                "http_method": http_method.value,
                "http_route": http_route,
                "action": action,
                "params_normalizados": json.dumps(
                    params_normalizados,
                    ensure_ascii=False,
                    separators=(",", ":"),
                ),
                "is_transaction_intent": (
                    "1"
                    if http_method
                    in {
                        HTTPMethods.POST,
                        HTTPMethods.PUT,
                        HTTPMethods.PATCH,
                        HTTPMethods.DELETE,
                    }
                    else "0"
                ),
            }

            if meta_human:
                event["meta_human"] = meta_human

            if request_raw_truncated:
                event["request_raw_truncated"] = request_raw_truncated
            elif payload_ref:
                event["payload_ref"] = payload_ref

            msg_id = self.redis_client.xadd(
                self.stream_key,
                fields=event,
            )

            return msg_id
        except Exception as e:
            get_logger().error(f"Erro ao publicar evento request_started: {e}")

    def emit_request_finished(
        self,
        request_id: uuid.UUID,
        tenant_id: int,
        grupo_empresarial_id: uuid.UUID | None,
        http_status: int,
        duration_ms: int,
        tx_attempted: bool,
        area_atendimento_id: uuid.UUID | None = None,
        db_user: str | None = None,
        error_normalized: dict | None = None,
        error_code: str | None = None,
        error_message_short: str | None = None,
        error_fingerprint: str | None = None,
        request_json: str | None = None,
        is_transaction_intent: bool | None = None,
    ) -> str:
        """
        Publica um evento 'request_finished' no Redis Stream.

        Parâmetros:
        - request_id: ID da requisição
        - tenant_id: ID do tenant
        - grupo_empresarial_id: ID do grupo empresarial
        - area_atendimento_id: ID da area de atendimento
        - db_user: Nome do usuário do banco
        - http_status: Código HTTP da resposta
        - duration_ms: Duração total da requisição
        - tx_attempted: Flag indicando tentativa de transação
        - error_normalized: Objeto com error_code, error_message_short e error_fingerprint
        - error_code: Código normalizado do erro
        - error_message_short: Mensagem curta do erro
        - error_fingerprint: Fingerprint do erro
        - request_json: JSON da requisição (somente em casos de erro)
        - is_transaction_intent: Flag de intenção transacional
        """

        try:
            now = datetime.now(timezone.utc)

            event = {
                "event_id": str(uuid.uuid4()),
                "event_ts_utc": now.isoformat(timespec="milliseconds"),
                "event_type": "request_finished",
                "request_id": str(request_id),
                "tenant_id": str(tenant_id),
                "grupo_empresarial_id": (
                    "" if grupo_empresarial_id is None else str(grupo_empresarial_id)
                ),
                "area_atendimento_id": (
                    "" if area_atendimento_id is None else str(area_atendimento_id)
                ),
                "db_user": "" if db_user is None else db_user,
                "http_status": str(http_status),
                "duration_ms": str(duration_ms),
                "tx_attempted": "1" if tx_attempted else "0",
            }

            if is_transaction_intent is not None:
                event["is_transaction_intent"] = "1" if is_transaction_intent else "0"

            if error_normalized is None and (
                error_code or error_message_short or error_fingerprint
            ):
                error_normalized = {
                    "error_code": error_code or "",
                    "error_message_short": error_message_short or "",
                    "error_fingerprint": error_fingerprint or "",
                }

            if error_normalized is not None:
                event["error_normalized"] = json.dumps(
                    error_normalized,
                    ensure_ascii=False,
                    separators=(",", ":"),
                )

            if error_code:
                event["error_code"] = str(error_code)
            if error_message_short:
                event["error_message_short"] = error_message_short
            if error_fingerprint:
                event["error_fingerprint"] = error_fingerprint

            if request_json:
                event["request_json"] = request_json

            msg_id = self.redis_client.xadd(
                self.stream_key,
                fields=event,
            )

            return msg_id
        except Exception as e:
            get_logger().error(f"Erro ao publicar evento request_finished: {e}")



from typing import Any, Dict, Set


def build_fields_tree(dto: Any) -> Dict[str, Set[str]]:
    root_fields: Set[str] = set()
    dto_dict = getattr(dto, "__dict__", {})
    for attr in (
        "fields_map",
        "list_fields_map",
        "object_fields_map",
        "one_to_one_fields_map",
        "aggregator_fields_map",
        "sql_join_fields_map",
        "left_join_fields_map",
    ):
        fields_map = getattr(dto, attr, None)
        if isinstance(fields_map, dict):
            for key in fields_map.keys():
                if key in dto_dict:
                    root_fields.add(key)
    return {"root": root_fields}


def build_expands_tree(dto: Any) -> Dict[str, Set[str]]:
    expands_root: Set[str] = set()
    one_to_one_fields = getattr(dto, "one_to_one_fields_map", None)
    if isinstance(one_to_one_fields, dict):
        expands_root |= set(one_to_one_fields.keys())
    return {"root": expands_root}


def convert_dto_full(dto: Any) -> Any:
    if hasattr(dto, "convert_to_dict") and callable(dto.convert_to_dict):
        fields_tree = build_fields_tree(dto)
        expands_tree = build_expands_tree(dto)
        return dto.convert_to_dict(fields_tree, expands_tree)
    return dto



from typing import Dict, Tuple

from nsj_audit_lib.settings import DATABASE_DRIVER, get_logger


class AuditDBModelUtil:
    _COLUMN_TYPES: Dict[str, str] = {
        "outbox_id": "uuid",
        "outbox_seq": "bigint",
        "created_at": "timestamptz",
        "tenant_id": "varchar(150)",
        "grupo_empresarial_id": "varchar(150)",
        "area_atendimento_id": "varchar(150)",
        "request_id": "uuid",
        "user_id": "text",
        "subject_user_id": "text",
        "session_id": "text",
        "action": "text",
        "resource_type": "text",
        "resource_id": "text",
        "params_normalizados": "jsonb",
        "commit_json": "jsonb",
        "payload_ref": "text",
        "payload_sha256": "text",
        "schema_version": "integer",
    }

    _CREATE_COLUMNS: Tuple[str, ...] = (
        "outbox_id uuid PRIMARY KEY",
        "outbox_seq bigserial",
        "created_at timestamptz NOT NULL DEFAULT now()",
        "tenant_id varchar(150) NULL",
        "grupo_empresarial_id varchar(150) NULL",
        "area_atendimento_id varchar(150) NULL",
        "request_id uuid NOT NULL",
        "user_id text NOT NULL",
        "subject_user_id text NULL",
        "session_id text NULL",
        "action text NOT NULL",
        "resource_type text NOT NULL",
        "resource_id text NULL",
        "params_normalizados jsonb NOT NULL DEFAULT '{}'::jsonb",
        "commit_json jsonb NULL",
        "payload_ref text NULL",
        "payload_sha256 text NULL",
        "schema_version integer NOT NULL DEFAULT 1",
    )

    _INDEXES: Tuple[str, ...] = (
        "CREATE UNIQUE INDEX IF NOT EXISTS ux_audit_outbox_seq ON audit_outbox (outbox_seq)",
        "CREATE INDEX IF NOT EXISTS ix_outbox_tenant_ge_request ON audit_outbox (tenant_id, grupo_empresarial_id, request_id)",
        "CREATE INDEX IF NOT EXISTS ix_outbox_tenant_ge_created ON audit_outbox (tenant_id, grupo_empresarial_id, created_at)",
        "CREATE INDEX IF NOT EXISTS ix_outbox_tenant_ge_seq ON audit_outbox (tenant_id, grupo_empresarial_id, outbox_seq)",
    )

    _TYPE_MATCHERS: Dict[str, Tuple[str, ...]] = {
        "uuid": ("uuid",),
        "bigint": ("bigint", "int8"),
        "integer": ("integer", "int4"),
        "text": ("text",),
        "varchar(150)": ("character varying", "varchar"),
        "jsonb": ("jsonb",),
        "timestamptz": ("timestamp with time zone", "timestamptz"),
    }
    _NULLABLE_COLUMNS = {"tenant_id", "grupo_empresarial_id", "area_atendimento_id"}

    @classmethod
    def _is_postgres(cls, db=None) -> bool:
        detected = cls._detect_db_driver(db)
        if detected is not None:
            return detected.startswith("postgres")
        return DATABASE_DRIVER.lower().startswith("postgres")

    @classmethod
    def is_postgres(cls, db=None) -> bool:
        return cls._is_postgres(db)

    @classmethod
    def _detect_db_driver(cls, db) -> str | None:
        if db is None:
            return None

        # Direct string driver on adapter
        for attr in (
            "driver",
            "db_driver",
            "database_driver",
            "dialect_name",
            "db_dialect",
        ):
            value = getattr(db, attr, None)
            if isinstance(value, str) and value:
                return value.lower()

        candidates = [db]
        for attr in (
            "_db",
            "engine",
            "_engine",
            "connection",
            "_connection",
            "conn",
            "_conn",
        ):
            obj = getattr(db, attr, None)
            if obj is not None:
                candidates.append(obj)

        for obj in candidates:
            # SQLAlchemy engine/connection: dialect.name
            dialect = getattr(obj, "dialect", None)
            if dialect is not None:
                name = getattr(dialect, "name", None)
                if isinstance(name, str) and name:
                    return name.lower()
                driver = getattr(dialect, "driver", None)
                if isinstance(driver, str) and driver:
                    return driver.lower()

            url = getattr(obj, "url", None)
            if url is not None:
                drivername = getattr(url, "drivername", None)
                if isinstance(drivername, str) and drivername:
                    return drivername.lower()
                get_backend_name = getattr(url, "get_backend_name", None)
                if callable(get_backend_name):
                    try:
                        backend = get_backend_name()
                    except Exception:
                        backend = None
                    if isinstance(backend, str) and backend:
                        return backend.lower()

            name = getattr(obj, "name", None)
            if isinstance(name, str) and name:
                return name.lower()

        return None

    @staticmethod
    def is_audit_outbox_model_error(exc: Exception) -> bool:
        message = str(exc).lower()
        if "audit_outbox" not in message:
            return False
        return any(
            token in message
            for token in (
                "does not exist",
                "undefined",
                "column",
                "relation",
                "datatype",
                "type",
            )
        )

    @classmethod
    def ensure_audit_outbox_schema(cls, db) -> None:
        if not cls._is_postgres(db):
            return

        if not cls._table_exists(db):
            cls._create_table(db)
        else:
            cls._ensure_columns(db)
            cls._ensure_types(db)
            cls._ensure_nullability(db)

        cls._ensure_indexes(db)

    @classmethod
    def _table_exists(cls, db) -> bool:
        row = db.execute_query_first_result(
            """
            SELECT 1
              FROM information_schema.tables
             WHERE table_schema = 'public'
               AND table_name = 'audit_outbox'
            """
        )
        return row is not None

    @classmethod
    def _create_table(cls, db) -> None:
        columns_sql = ",\n    ".join(cls._CREATE_COLUMNS)
        sql = f"""
        CREATE TABLE IF NOT EXISTS audit_outbox (
            {columns_sql}
        );
        """
        db.execute(sql)

    @classmethod
    def _ensure_columns(cls, db) -> None:
        rows = db.execute_query(
            """
            SELECT column_name
              FROM information_schema.columns
             WHERE table_schema = 'public'
               AND table_name = 'audit_outbox'
            """
        )
        existing = {row["column_name"] for row in rows}

        for column_name, column_type in cls._COLUMN_TYPES.items():
            if column_name in existing:
                continue
            db.execute(
                f"ALTER TABLE audit_outbox ADD COLUMN {column_name} {column_type}"
            )

    @classmethod
    def _ensure_types(cls, db) -> None:
        rows = db.execute_query(
            """
            SELECT column_name, data_type, udt_name
              FROM information_schema.columns
             WHERE table_schema = 'public'
               AND table_name = 'audit_outbox'
            """
        )
        columns = {
            row["column_name"]: (row["data_type"], row["udt_name"]) for row in rows
        }

        for column_name, column_type in cls._COLUMN_TYPES.items():
            current = columns.get(column_name)
            if current is None:
                continue

            data_type, udt_name = current
            valid_types = cls._TYPE_MATCHERS.get(column_type, (column_type,))
            if data_type in valid_types or udt_name in valid_types:
                continue

            db.execute(
                f"ALTER TABLE audit_outbox ALTER COLUMN {column_name} "
                f"TYPE {column_type} USING {column_name}::{column_type}"
            )

    @classmethod
    def _ensure_nullability(cls, db) -> None:
        rows = db.execute_query(
            """
            SELECT column_name, is_nullable
              FROM information_schema.columns
             WHERE table_schema = 'public'
               AND table_name = 'audit_outbox'
            """
        )
        nullable_map = {row["column_name"]: row["is_nullable"] for row in rows}

        for column_name in cls._NULLABLE_COLUMNS:
            if nullable_map.get(column_name) == "NO":
                db.execute(
                    f"ALTER TABLE audit_outbox ALTER COLUMN {column_name} DROP NOT NULL"
                )

    @classmethod
    def _ensure_indexes(cls, db) -> None:
        for sql in cls._INDEXES:
            try:
                db.execute(sql)
            except Exception as exc:
                get_logger().warning(f"Erro criando indice de audit_outbox: {exc}")



import base64
import json

from flask import g, request

from nsj_audit_lib.settings import DATABASE_USER


def _decode_jwt_payload(token: str):
    parts = token.split(".")
    if len(parts) < 2:
        return None

    payload_b64 = parts[1]
    padding = "=" * (-len(payload_b64) % 4)
    try:
        payload_json = base64.urlsafe_b64decode(payload_b64 + padding).decode("utf-8")
    except (ValueError, UnicodeDecodeError):
        return None

    try:
        payload = json.loads(payload_json)
    except json.JSONDecodeError:
        return None

    return payload if isinstance(payload, dict) else None


def get_db_user():
    """
    Resolve o usuário do banco para auditoria.

    Ordem de tentativa:
    1) g.external_database["user"] quando definido.
    2) fallback para DATABASE_USER (settings.py).
    """
    external_database = getattr(g, "external_database", None)
    if isinstance(external_database, dict):
        external_user = external_database.get("user")
        if external_user:
            return external_user

    return DATABASE_USER


def get_actor_user_id():
    """
    Resolve o identificador do usuário para auditoria.

    Ordem de tentativa:
    1) g.profile["email"] quando definido.
    2) Authorization Bearer: extrai "email" do payload JWT.
    3) Authorization Basic: usa o usuário antes do ":" após base64.
       Retorna f"sistema2_{id_sistema}".
    4) X-API-Key ou apiKey: trata como JWT e usa "sistema" -> "id_sistema".
       Retorna f"sistema_{id_sistema}".
    5) fallback para o usuário do banco (get_db_user).
    """
    profile = getattr(g, "profile", None)
    if isinstance(profile, dict):
        email = profile.get("email")
        if email:
            return email

    authorization = request.headers.get("Authorization")
    if authorization:
        scheme, _, value = authorization.partition(" ")
        if value:
            scheme_lower = scheme.lower()
            if scheme_lower == "bearer":
                payload = _decode_jwt_payload(value)
                if payload and payload.get("email"):
                    return payload.get("email")
            elif scheme_lower == "basic":
                try:
                    decoded = base64.b64decode(value).decode("utf-8")
                except (ValueError, UnicodeDecodeError):
                    decoded = ""
                if decoded:
                    return f"sistema2_{decoded.split(':', 1)[0]}"

    api_key = request.headers.get("X-API-Key") or request.headers.get("apiKey")
    if api_key:
        payload = _decode_jwt_payload(api_key)
        if payload:
            sistema = payload.get("sistema")
            if isinstance(sistema, dict):
                id_sistema = sistema.get("id_sistema")
            else:
                id_sistema = sistema
            if id_sistema is not None:
                return f"sistema_{id_sistema}"

    return get_db_user()



import hashlib
import json

from typing import Any


def parse_status_code(status_value: Any) -> int | None:
    if isinstance(status_value, int):
        return status_value
    if isinstance(status_value, str):
        for part in status_value.split():
            if part.isdigit():
                return int(part)
    return None


def parse_bool(value: Any) -> bool | None:
    if isinstance(value, bool):
        return value
    if isinstance(value, (int, float)):
        return bool(value)
    if isinstance(value, str):
        return value.strip().lower() in {"1", "true", "t", "yes", "y"}
    return None


def unpack_response(response: Any) -> tuple[Any, int]:
    if response is None:
        return None, 500

    body = response
    http_status = None

    if isinstance(response, tuple):
        if response:
            body = response[0]
        if len(response) > 1:
            http_status = parse_status_code(response[1])
        if http_status is None and len(response) > 2:
            http_status = parse_status_code(response[2])
    elif hasattr(response, "status_code"):
        http_status = getattr(response, "status_code", None)
        if hasattr(response, "get_data"):
            body = response.get_data(as_text=True)

    if http_status is None:
        http_status = 200

    return body, http_status


def normalize_response_body(body: Any) -> tuple[str, Any | None]:
    body_json = None
    if isinstance(body, (dict, list)):
        body_json = body
        body_text = json.dumps(body, ensure_ascii=False, separators=(",", ":"))
    elif isinstance(body, bytes):
        body_text = body.decode("utf-8", errors="replace")
    elif isinstance(body, str):
        body_text = body
    elif body is None:
        body_text = ""
    else:
        body_text = str(body)

    if body_json is None and body_text:
        trimmed = body_text.lstrip()
        if trimmed.startswith("{") or trimmed.startswith("["):
            try:
                body_json = json.loads(body_text)
            except json.JSONDecodeError:
                body_json = None

    return body_text, body_json


def mask_sensitive_data(value: Any) -> Any:
    if isinstance(value, dict):
        sanitized = {}
        for key, item in value.items():
            key_text = str(key).lower()
            if any(token in key_text for token in ("password", "senha", "pass", "secret")):
                sanitized[key] = "******"
            else:
                sanitized[key] = mask_sensitive_data(item)
        return sanitized
    if isinstance(value, list):
        return [mask_sensitive_data(item) for item in value]
    return value


def sanitize_payload(raw_body: str) -> str:
    trimmed = raw_body.lstrip()
    if trimmed.startswith("{") or trimmed.startswith("["):
        try:
            parsed = json.loads(raw_body)
        except json.JSONDecodeError:
            return raw_body
        sanitized = mask_sensitive_data(parsed)
        return json.dumps(sanitized, ensure_ascii=False, separators=(",", ":"))
    return raw_body


def extract_error_payload(body_json: Any) -> dict | None:
    if isinstance(body_json, list):
        for item in body_json:
            if isinstance(item, dict):
                return item
        return None
    if isinstance(body_json, dict):
        return body_json
    return None


def hash_message(message: str) -> str:
    return hashlib.sha256(message.encode("utf-8")).hexdigest()


def extract_error_info(
    http_status: int,
    body_json: Any,
    body_text: str,
) -> tuple[str, str, str]:
    if http_status < 400:
        return "", "", ""

    # TODO Refatorar para MULTI STATUS no futuro
    error_payload = extract_error_payload(body_json)
    error_code = ""
    error_message_short = ""
    error_fingerprint = ""

    if isinstance(error_payload, dict):
        error_code = error_payload.get("code") or error_payload.get("error_code") or ""
        message = (
            error_payload.get("message")
            or error_payload.get("error_message_short")
            or error_payload.get("mensagem")
            or error_payload.get("error")
            or ""
        )
        error_message_short = message
        payload_text = json.dumps(
            error_payload,
            ensure_ascii=False,
            separators=(",", ":"),
            sort_keys=True,
        )
        error_fingerprint = hash_message(payload_text)

    if not error_message_short and body_text:
        error_message_short = body_text.strip()

    if error_message_short:
        error_message_short = error_message_short[:255]

    if not error_fingerprint:
        if error_code:
            error_fingerprint = str(error_code)
        elif error_message_short:
            error_fingerprint = hash_message(error_message_short)

    return str(error_code or ""), error_message_short, error_fingerprint



import uuid

from typing import Any

from nsj_gcf_utils.db_adapter2 import DBAdapter2
from nsj_gcf_utils.json_util import json_dumps, convert_to_dumps
from nsj_audit_lib.settings import get_logger
from nsj_audit_lib.util.audit_dbmodel_util import AuditDBModelUtil


class DAOBaseAudit:
    def __init__(self, db_adapter: DBAdapter2) -> None:
        self._db = db_adapter

    def insert_outbox(self, payload: dict[str, Any]) -> None:
        outbox_payload = dict(payload or {})
        outbox_payload.setdefault("outbox_id", str(uuid.uuid4()))
        outbox_payload.setdefault("schema_version", 1)

        outbox_payload["params_normalizados"] = self._normalize_json_field(
            outbox_payload.get("params_normalizados") or {}
        )
        outbox_payload["commit_json"] = self._normalize_json_field(
            outbox_payload.get("commit_json")
        )

        outbox_payload = convert_to_dumps(outbox_payload)

        sql = """
        INSERT INTO audit_outbox (
            outbox_id,
            tenant_id,
            grupo_empresarial_id,
            area_atendimento_id,
            request_id,
            user_id,
            subject_user_id,
            session_id,
            action,
            resource_type,
            resource_id,
            params_normalizados,
            commit_json,
            payload_ref,
            payload_sha256,
            schema_version
        ) VALUES (
            :outbox_id,
            :tenant_id,
            :grupo_empresarial_id,
            :area_atendimento_id,
            :request_id,
            :user_id,
            :subject_user_id,
            :session_id,
            :action,
            :resource_type,
            :resource_id,
            :params_normalizados,
            :commit_json,
            :payload_ref,
            :payload_sha256,
            :schema_version
        )
        """

        savepoint_set = False
        logger = get_logger()
        logger.debug(
            "audit_outbox insert start: db=%s driver=%s in_tx=%s",
            type(self._db).__name__,
            AuditDBModelUtil._detect_db_driver(self._db),
            self._db.in_transaction(),
        )
        try:
            if self._db.in_transaction():
                self._db.execute("SAVEPOINT audit_outbox_sp")
                savepoint_set = True

            rowcount, _ = self._db.execute(sql, **outbox_payload)
            if rowcount <= 0:
                raise Exception("Erro inserindo audit_outbox no banco de dados.")
        except Exception as exc:
            logger.debug("audit_outbox insert failed: %s", exc)
            if savepoint_set:
                try:
                    self._db.execute("ROLLBACK TO SAVEPOINT audit_outbox_sp")
                except Exception:
                    get_logger().warning(
                        "Erro desconhecido ao fazer rollback do savepoint de audit_outbox."
                    )
                    pass

            is_postgres = AuditDBModelUtil.is_postgres(self._db)
            is_model_error = AuditDBModelUtil.is_audit_outbox_model_error(exc)
            logger.debug(
                "audit_outbox repair check: is_postgres=%s is_model_error=%s",
                is_postgres,
                is_model_error,
            )
            if is_postgres and is_model_error:
                self._repair_schema()
                retry_savepoint_set = False
                try:
                    if self._db.in_transaction():
                        self._db.execute("SAVEPOINT audit_outbox_sp_retry")
                        retry_savepoint_set = True
                    rowcount, _ = self._db.execute(sql, **outbox_payload)
                    if rowcount <= 0:
                        raise Exception(
                            "Erro inserindo audit_outbox no banco de dados."
                        )
                    return
                except Exception as retry_exc:
                    if retry_savepoint_set:
                        try:
                            self._db.execute(
                                "ROLLBACK TO SAVEPOINT audit_outbox_sp_retry"
                            )
                        except Exception:
                            get_logger().warning(
                                "Erro desconhecido ao fazer rollback do savepoint de audit_outbox (retry)."
                            )
                    raise Exception(
                        "Erro inserindo audit_outbox no banco de dados."
                    ) from retry_exc

            raise

    @staticmethod
    def _normalize_json_field(value: Any) -> Any:
        if value is None:
            return None
        if isinstance(value, str):
            return value
        return json_dumps(value, ensure_ascii=False)

    def _repair_schema(self) -> None:
        logger = get_logger()
        logger.debug(
            "audit_outbox repair start: db=%s driver=%s",
            type(self._db).__name__,
            AuditDBModelUtil._detect_db_driver(self._db),
        )

        AuditDBModelUtil.ensure_audit_outbox_schema(self._db)
        return



import hashlib
import json
import uuid

from typing import Any, Dict

from flask import g, has_request_context, request

from nsj_gcf_utils.db_adapter2 import DBAdapter2

from nsj_audit_lib.dao.dao_base_audit import DAOBaseAudit
from nsj_audit_lib.settings import AUDIT_OUTBOX_TRANSACTION
from nsj_audit_lib.util.audit_ids_util import (
    retrieve_area_atendimento_id,
    retrieve_grupo_empresarial_id,
    retrieve_tenant_id,
)
from nsj_audit_lib.util.audit_dto_util import convert_dto_full
from nsj_audit_lib.util.user_audit_util import get_actor_user_id
from nsj_audit_lib.util.util_normaliza_parametros import get_params_normalizados


class AuditService:

    def __init__(self, db: DBAdapter2) -> None:
        self.audit_dao = DAOBaseAudit(db_adapter=db)

    def should_record_audit_outbox(self) -> bool:
        if not AUDIT_OUTBOX_TRANSACTION:
            return False
        if not has_request_context():
            return False
        return getattr(g, "request_id", None) is not None

    def record_audit_outbox(
        self,
        action: str,
        dto: Any | None,
        resource_id: Any | None,
        params_normalizados: Dict[str, Any] | None = None,
        commit_json: Any | None = None,
        old_dto: Any | None = None,
        route_resource_id: Any | None = None,
    ) -> None:
        if not AUDIT_OUTBOX_TRANSACTION:
            return
        if not has_request_context():
            return

        request_id = getattr(g, "request_id", None)
        if request_id is None:
            return

        audit_payload = self._build_audit_payload(
            action=action,
            dto=dto,
            resource_id=resource_id,
            params_normalizados=params_normalizados,
            commit_json=commit_json,
            old_dto=old_dto,
            route_resource_id=route_resource_id,
        )
        if audit_payload is None:
            return

        self.audit_dao.insert_outbox(audit_payload)

    def _build_audit_payload(
        self,
        action: str,
        dto: Any | None,
        resource_id: Any | None,
        params_normalizados: Dict[str, Any] | None,
        commit_json: Any | None,
        old_dto: Any | None,
        route_resource_id: Any | None,
    ) -> Dict[str, Any] | None:
        request_id = getattr(g, "request_id", None)
        if request_id is None:
            return None

        query_args = request.args
        body = request.get_json(force=True, silent=True)

        tenant_id = getattr(g, "audit_tenant_id", None) or retrieve_tenant_id(
            None,
            query_args=query_args,
            body=body,
        )
        grupo_empresarial_id = getattr(
            g, "audit_grupo_empresarial_id", None
        ) or retrieve_grupo_empresarial_id(
            None,
            query_args=query_args,
            body=body,
        )
        area_atendimento_id = getattr(
            g, "audit_area_atendimento_id", None
        ) or retrieve_area_atendimento_id(
            None,
            query_args=query_args,
            body=body,
        )

        user_id = get_actor_user_id()
        subject_user_id = self._resolve_subject_user_id()
        session_id = self._resolve_session_id()

        params_normalizados = params_normalizados or getattr(
            g, "audit_params_normalizados", None
        )
        if params_normalizados is None:
            params_normalizados = get_params_normalizados(
                query_args=query_args,
                body=body,
                dto_class=getattr(self, "_dto_class", None),
                path_args=getattr(request, "view_args", None),
            )

        resource_id = resource_id or self._resolve_resource_id(dto)
        if commit_json is None:
            commit_json = self._build_commit_json(old_dto, dto, resource_id)
        commit_json = self._add_table_name_to_commit_json(commit_json)

        resource_type = self._resolve_resource_type(route_resource_id)
        payload_sha256 = self._build_payload_sha256(body, query_args)

        return {
            "tenant_id": self._normalize_audit_id_value(tenant_id),
            "grupo_empresarial_id": self._normalize_audit_id_value(
                grupo_empresarial_id
            ),
            "area_atendimento_id": self._normalize_audit_id_value(area_atendimento_id),
            "request_id": request_id,
            "user_id": user_id,
            "subject_user_id": subject_user_id,
            "session_id": session_id,
            "action": action,
            "resource_type": resource_type,
            "resource_id": resource_id,
            "params_normalizados": params_normalizados or {},
            "commit_json": commit_json,
            "payload_ref": None,
            "payload_sha256": payload_sha256,
            "schema_version": 1,
        }

    def _add_table_name_to_commit_json(self, commit_json: Any | None) -> Any | None:
        if not isinstance(commit_json, dict):
            return commit_json

        table_name = None
        entity_class = getattr(self, "_entity_class", None)
        if entity_class is not None:
            try:
                table_name = entity_class().get_table_name()
            except Exception:
                table_name = None

        if table_name:
            commit_json.setdefault("table_name", table_name)

        return commit_json

    @staticmethod
    def _normalize_audit_id_value(value: Any) -> Any:
        if isinstance(value, uuid.UUID):
            return str(value)
        return value

    def _build_commit_json(
        self,
        old_dto: Any | None,
        new_dto: Any | None,
        resource_id: Any | None,
    ) -> Any:
        if old_dto is not None and new_dto is not None:
            old_values = convert_dto_full(old_dto)
            new_values = convert_dto_full(new_dto)
            if not isinstance(old_values, dict) or not isinstance(new_values, dict):
                return new_values
            diff = self._diff_values(old_values, new_values)
            return diff if diff else None

        if new_dto is None:
            if resource_id is None:
                return None
            return {"id": resource_id}

        return convert_dto_full(new_dto)

    @staticmethod
    def _diff_values(
        old_values: Dict[str, Any],
        new_values: Dict[str, Any],
    ) -> Dict[str, Any]:
        diff: Dict[str, Any] = {}
        for key in set(old_values.keys()) | set(new_values.keys()):
            old_value = old_values.get(key)
            new_value = new_values.get(key)
            if old_value != new_value:
                diff[key] = {"old": old_value, "new": new_value}
        return diff

    @staticmethod
    def _build_payload_sha256(body: Any, query_args: Any) -> str | None:
        if body:
            try:
                payload = json.dumps(
                    body,
                    ensure_ascii=False,
                    separators=(",", ":"),
                    sort_keys=True,
                )
            except TypeError:
                payload = str(body)
        else:
            if hasattr(query_args, "to_dict"):
                payload_source = query_args.to_dict(flat=True)
            else:
                payload_source = dict(query_args or {})
            payload = json.dumps(
                payload_source,
                ensure_ascii=False,
                separators=(",", ":"),
                sort_keys=True,
            )

        if not payload:
            return None

        return hashlib.sha256(payload.encode("utf-8")).hexdigest()

    def _resolve_resource_type(self, resource_id: Any | None) -> str:
        if has_request_context():
            resource = self._resource_from_path(request.path or "", resource_id)
            if resource:
                return resource

        dto_class = getattr(self, "_dto_class", None)
        if dto_class is not None:
            return dto_class.__name__

        return ""

    def _resolve_resource_id(self, dto: Any | None) -> Any | None:
        if dto is None:
            return None
        pk_field = getattr(dto, "pk_field", None)
        if pk_field and hasattr(dto, pk_field):
            return getattr(dto, pk_field)
        return None

    @staticmethod
    def _resource_from_path(path: str, resource_id: Any | None) -> str | None:
        segments = [segment for segment in path.split("/") if segment]
        if not segments:
            return None

        if resource_id is not None:
            if len(segments) > 1:
                return segments[-2]
            return None

        return segments[-1]

    @staticmethod
    def _resolve_subject_user_id() -> None:
        return None

    def _resolve_session_id(self) -> str | None:
        profile = getattr(g, "profile", None)
        if isinstance(profile, dict):
            for key in ("session_id", "session", "sessao"):
                value = profile.get(key)
                if value:
                    return str(value)
        return None
